#vider la mmoire
rm(list=ls())

#changement de rpertoire
setwd("D:/DataMining/Databases_for_mining/dataset_for_soft_dev_and_comparison/stacking")

#charger le fichier
spam.all <- read.table("spam_stacking.txt", header = T, sep="\t", dec=".")

#vrification
print(str(spam.all))

#index pour partition
set.seed(1)
idx <- sample(1:nrow(spam.all),1601,replace=FALSE)

#apprentissage
spam.train <- spam.all[idx,]
print(dim(spam.train))

#test
spam.test <- spam.all[-idx,]
print(dim(spam.test))

#calcul des moyennes sur l'chantillon d'apprentissage
mean.train <- sapply(spam.train[-ncol(spam.train)],mean)
print(mean.train)

#idem pour l'cart-type
sd.train <- sapply(spam.train[-ncol(spam.train)],sd)
print(sd.train)

#centrage-rduction de l'chantillon d'apprentissage
spam.train.cr <- data.frame(scale(spam.train[-ncol(spam.train)],center=mean.train,scale=sd.train))

#vrification
print(colMeans(spam.train.cr))

#adjonction de la variable cible
spam.train.cr$spam <- spam.train$spam

#centrage et rduction de l'chantillon test
#avec les paramtres (moyenne, cart-type) calculs sur l'chantillon d'apprentissage
spam.test.cr <- data.frame(scale(spam.test[-ncol(spam.test)], center = mean.train, scale = sd.train))

#moyenne pas forcment nulle - c'est tout  fait normal
print(colMeans(spam.test.cr))

#adjonction de la cible
spam.test.cr$spam <- spam.test$spam

#########################
## Modles individuels ##
#########################

#RPART pour rpart
library(rpart)
arbre.seul <- rpart(spam ~ ., data = spam.train.cr, control = rpart.control(cp = 0))
print(arbre.seul)

#prdiction des arbres
pred.arbre <- predict(arbre.seul,newdata=spam.test.cr,type="class")
print(table(pred.arbre))

#CARET pour valuation
library(caret)

#taux de reconnaissance
caret::confusionMatrix(data=pred.arbre,reference=spam.test.cr$spam,positive="yes")

#probabilit d'appartenance aux classes
proba.arbre <- predict(arbre.seul,newdata=spam.test.cr,type="prob")
print(head(proba.arbre))

#MASS pour lda
library(MASS)
adl.seul <- lda(spam ~ ., data = spam.train.cr)
print(adl.seul)

#prdiction ADL
pred.adl <- predict(adl.seul,newdata=spam.test.cr)$class
print(table(pred.adl))

#perfomances
caret::confusionMatrix(data=pred.adl,reference=spam.test.cr$spam)$overall["Accuracy"]

#probabilits d'appartenance aux classes
proba.adl <- predict(adl.seul,newdata=spam.test.cr)$posterior
head(proba.adl)

#KERNLAB pour SVM RBF
library(kernlab)
svm.seul <- ksvm(spam ~ ., data = spam.train.cr, kernel = "rbfdot", C = 1.0, kpar = list(sigma = 0.1), prob.model = TRUE)
svm.seul

#prdiction svm
pred.svm <- predict(svm.seul,newdata=spam.test.cr,type="response")
print(table(pred.svm))

#performances
caret::confusionMatrix(data=pred.svm,reference=spam.test.cr$spam)$overall["Accuracy"]

#probas d'appartenance aux classes
proba.svm <- predict(svm.seul,newdata=spam.test.cr,type="probabilities")
print(head(proba.svm))

#######################################
## Vote "hard" - Sur les prdictions ##
#######################################

#prdiction regroupes
pred.all <- data.frame(arbre=as.numeric(pred.arbre=="yes"),adl=as.numeric(pred.adl=="yes"),svm=as.numeric(pred.svm=="yes"))
print(head(pred.all))

#corrlation entre les prdictions
cor(pred.all)

#mtamodle - vote "hard"
pred.hard <- factor(ifelse(rowSums(pred.all) > 1.5,"yes","no"))
print(table(pred.hard))                       

#performances du vote "hard"
caret::confusionMatrix(data=pred.hard,reference=spam.test.cr$spam)$overall["Accuracy"]

########################################
## Vote "soft" - Sur les probabilits ##
########################################

#probas. "yes" regroupes
proba.all <- data.frame(arbre=proba.arbre[,"yes"],adl=proba.adl[,"yes"],svm=proba.svm[,"yes"])
print(head(proba.all))

#corrlations
cor(proba.all)

#scatter croiss
pairs(proba.all)

#calibration - comparabilit des probas
summary(proba.all)

#graphiquement avec des boxplot
boxplot(proba.all)

#voyons le vote "soft"
pred.soft <- factor(ifelse(rowSums(proba.all) > 1.5, "yes", "no"))
print(table(pred.soft))

#performances du vote "soft"
caret::confusionMatrix(data=pred.soft,reference=spam.test.cr$spam)$overall["Accuracy"]

#cohrence entre les deux prdictions
table(pred.hard,pred.soft)

####################################
## Stacking - Sur les prdictions ##
####################################

#nombre de folds
nfolds <- 5

#cration des 5 blocs
set.seed(100)
index <- 1 + order(runif(nrow(spam.train))) %% nfolds
print(head(index,20))

#nombre d'observations par bloc
print(table(index))

#matrice pour collecter les rsultats de prdictions en validation croise
mPredCv <- matrix(0,nrow=nrow(spam.train),ncol=3)

#nommer les colonnes
colnames(mPredCv) <- c("arbre","adl","svm")

#sessions apprentissage-test
for (k in 1:nfolds){
  ###########################
  #### ARBRE DE DECISION ####
  ###########################
  #apprentissage arbre sur les autres que le fold numro k
  arbreCv <- rpart(spam ~ ., data = spam.train.cr[index!=k,], control = rpart.control(cp = 0))
  #prdiction arbre sur le fold numro k
  predArbreCV <- as.numeric(predict(arbreCv,newdata=spam.train.cr[index==k,],type="class")=="yes")
  #intgration de la prdiction dans la matrice
  mPredCv[index==k,"arbre"] <- predArbreCV
  ###############################
  #### ANALYSE DISCRIMINANTE ####
  ###############################
  #apprentissage
  adlCv <- lda(spam ~ ., data = spam.train.cr[index!=k,])
  #prdiction
  predAdlCv <- as.numeric(predict(adlCv,newdata=spam.train.cr[index==k,])$class == "yes")
  #rcupration
  mPredCv[index==k,"adl"] <- predAdlCv
  #################
  #### SVM RBF ####
  #################
  #apprentissage
  svmCv <- ksvm(spam ~ ., data = spam.train.cr[index!=k,], scale = FALSE, kernel = "rbfdot", C = 1.0, kpar = list(sigma = 0.1), prob.model = TRUE)
  #prdiction
  predSvmCv <- as.numeric(predict(svmCv,newdata=spam.train.cr[index==k,],type="response")=="yes")
  #rcupration de la prdiction
  mPredCv[index==k,"svm"] <- predSvmCv
}

#vrification
print(head(mPredCv,20))

#adjonction de la variable cible
dataCv <- data.frame(cbind(spam.train.cr["spam"],data.frame(mPredCv)))
print(head(dataCv))

#modlisation stacking
modelStack <- glm(spam ~ ., data = dataCv, family="binomial")
print(modelStack)

#utilisation de ce modle pour pondrer les prdictions "hard"
#valeur du logit
logit.pred.Stack <- predict(modelStack,newdata = pred.all,type="link")
print(head(logit.pred.Stack))

#conversion en prdiction
pred.Stack <- factor(ifelse(logit.pred.Stack > 0, "yes", "no"))
print(table(pred.Stack))

#performances
caret::confusionMatrix(data=pred.Stack,reference=spam.test.cr$spam)$overall["Accuracy"]

#on peut envisager la mme chose mais en travaillant sur les probas !!!

####################
## Caret Ensemble ##
####################

#caret
library(caret)

#paramtre pour l'apprentissage - si on veut assembler, il faut absolument passer par une mthode de resampling
#ici nous utilisons la validation croise en 5 blocs
#et il faut aussi produire les probabilits d'affectation des classes
fitControl <- caret::trainControl(method="cv",number=5, classProbs = TRUE,savePredictions="all")

#caretEnsemble
library(caretEnsemble)

#liste des modles  empiler
#https://topepo.github.io/caret/available-models.html
arbre <- caretEnsemble::caretModelSpec(method="rpart",tuneGrid=data.frame(cp=0))
adl <- caretEnsemble::caretModelSpec(method="lda")
svmRbf <- caretEnsemble::caretModelSpec(method="svmRadial",tuneGrid=data.frame(C=1,sigma=0.1))

#lancement des mthodes en validation croise
set.seed(1)
modeles <- caretEnsemble::caretList(spam ~ ., data = spam.train.cr, trControl = fitControl, tuneList = list(arbre,adl,svmRbf))
print(modeles)

#collecter le dtail des rsultats en validation croise
resultatsCV <- caret::resamples(modeles)
print(resultatsCV)

#affichage
print(summary(resultatsCV))

#affichage graphique "intervalle de variation"
dotplot(resultatsCV)

#corrlation entre les rsulats dans les folds
#corrlation = 1 ne veut pas dire que les mthodes classent de la mme manire
#mais que les taux sont les mmes dans les folds
caret::modelCor(resultatsCV)

#taux de succs pour les folds
print(resultatsCV$values)

#vrification de la corrlation pour deux modles
cor(resultatsCV$values["rpart~Accuracy"],resultatsCV$values["lda~Accuracy"])

#affichage grahique crois des rsultats par folds
#"lattice" a t install et charg automatiquement
#en mme temps de caret
lattice::splom(resultatsCV)

#stacking via la glm
#trainControl(method="none") - pas (ncessairement) besoin de resampling ici
mStack <- caretEnsemble::caretStack(modeles,method="glm",trControl=trainControl(method="none"))
print(mStack)

#affichage dtaill du mtamodle (glm)
#visualisation des coefficients
summary(mStack)

#prdiction et valuation
pStack <- predict(mStack,newdata=spam.test.cr)
caret::confusionMatrix(spam.test.cr$spam,pStack)$overall["Accuracy"]

#############################
## H20 - Stacked Ensembles ##
#############################

#library
library(h2o)

#initialisation
h2o.init(nthreads = -1)

#transformer en un format reconnu par h2o - training set
h2oTrain <- as.h2o(spam.train.cr)

#idem pur le test set
h2oTest <- as.h2o(spam.test.cr)

#premier modle - 1 arbre via gradient boosting
h2o_arbre <- h2o.gbm(y="spam",training_frame=h2oTrain,ntrees=1,nfolds=5,keep_cross_validation_predictions = TRUE,seed=1)
print(h2o_arbre)

#ses performances en test
caret::confusionMatrix(spam.test.cr$spam,as.data.frame(h2o.predict(h2o_arbre,newdata=h2oTest))$predict)$overall["Accuracy"]

#second modle - naive bayes
h2o_nb <- h2o.naiveBayes(y="spam",training_frame=h2oTrain,nfolds=5,keep_cross_validation_predictions = TRUE,seed=1)
print(h2o_nb)

#perfs en test
caret::confusionMatrix(spam.test.cr$spam,as.data.frame(h2o.predict(h2o_nb,newdata=h2oTest))$predict)$overall["Accuracy"]

#perceptron  10 neurones dans la couche cache
h2o_mlp <- h2o.deeplearning(y="spam",training_frame=h2oTrain,hidden=c(10),nfolds=5,keep_cross_validation_predictions = TRUE,seed=1)
print(h2o_mlp)

#perfs en test
caret::confusionMatrix(spam.test.cr$spam,as.data.frame(h2o.predict(h2o_mlp,newdata=h2oTest))$predict)$overall["Accuracy"]

#modle empil
h2o_stack <- h2o.stackedEnsemble(y="spam",training_frame=h2oTrain, base_models = list(h2o_arbre,h2o_nb,h2o_mlp),metalearner_algorithm = "glm",seed=2019)
print(h2o_stack)

#ses perfromances
caret::confusionMatrix(spam.test.cr$spam,as.data.frame(h2o.predict(h2o_stack,newdata=h2oTest))$predict)$overall["Accuracy"]

#arrt
h2o.shutdown()

